/*
 * playlistaspectoverviewfilter.cpp
 *
 *  Created on: 20.04.2010
 *      Author: darkstar
 */

#include "playlistaspectoverviewfilter.h"
#include "sqlite3.h"
#include "debug.h"
#include "compathack.h"

PlayListAspectOverviewFilter::PlayListAspectOverviewFilter()
	:	QObject(),
	 	PlayListAspectExtension(),
	 	isDirty_(false),
		updatingGenres_(false),
		updatingArtists_(false),
		updatingAlbums_(false),
		selectedGenresFilter(""),
		selectedArtistsFilter(""),
		selectedAlbumsFilter("")
{

}

PlayListAspectOverviewFilter::~PlayListAspectOverviewFilter()
{

}

void PlayListAspectOverviewFilter::reset()
{

}

void PlayListAspectOverviewFilter::execute()
{
	DENTERMETHOD("PlayListAspectOverviewFilter::execute()");

	// if the overview widget is invisible and no genre, artist or album filter is set, we can just set
	// it into dirty state and thus speed up processing. Once the overview widget is shown again, we check for isDirty
	// in the showEvent method and update the lists' content accordingly...
	if (isActive() && genreList_.at(0)->selected && artistList_.at(0)->selected && albumList_.at(0)->selected)
	{
		isDirty_ = true;
		PlayListAspectExtension::execute();
	}
	else
	{
		isDirty_ = false;

		sqlite3_exec(playListAspect()->playList()->mediaDatabase()->db(),
			"DROP TABLE IF EXISTS overview;"
			"CREATE TEMPORARY TABLE overview AS "
				"SELECT DISTINCT artist, album, genre FROM " + inputViewSource() + ";", NULL, NULL, NULL);

		updateFromGenreList();
	}

	DEXITMETHOD("PlayListAspectOverviewFilter::execute()");
}

void PlayListAspectOverviewFilter::saveExtensionState(QStringList &dst)
{
	dst.clear();
	dst.append("genre:" + buildFilter(genreList_, tr("(Undefined genre)")));
	dst.append("artist:" + buildFilter(artistList_, tr("(Unknown artist)")));
	dst.append("album:" + buildFilter(albumList_, tr("(Unknown album)")));
}

void PlayListAspectOverviewFilter::loadExtensionState(QStringList &src)
{
	if (src.count() == 3)
	{
		selectedGenresFilter = src[0].mid(src[0].indexOf(":") + 1);
		selectedArtistsFilter = src[1].mid(src[1].indexOf(":") + 1);
		selectedAlbumsFilter = src[2].mid(src[2].indexOf(":") + 1);
		isDirty_ = true;
		execute();
	}
}

int PlayListAspectOverviewFilter::selectionCount(OverviewFilterList &list)
{
	int selectedCount = 0;

	for (int i = 1; i < list.count(); ++i)
		if (list.at(i)->selected)
			++selectedCount;

	return selectedCount;
}

QString PlayListAspectOverviewFilter::buildFilter(OverviewFilterList &list, const QString &textForUnknownEntry)
{
	QString filter = "";
	int selectedCount = 0;

	if (list.count() > 1 && !list.at(0)->selected)
		for (int i = 1; i < list.count(); ++i)
			if (list.at(i)->selected)
			{
				++selectedCount;
				if (list.at(i)->text == textForUnknownEntry)
					filter += (selectedCount > 1 ? QString(", ") : QString("")) + "NULL, \"\"";
				else
					filter += (selectedCount > 1 ? QString(", ") : QString("")) + "\"" + dbSafeString(list.at(i)->text) + "\"";
			}

	return filter;
}

void PlayListAspectOverviewFilter::createFilteredTable(const QString &selectedFilter, const QString &srctable, const QString &dsttable, const QString &what, const QString &wherecol)
{
	QString query =
		"DROP TABLE IF EXISTS " + dsttable + ";"
		"CREATE TEMPORARY TABLE " + dsttable + " AS "
			"SELECT DISTINCT " + what + " FROM " + srctable +
			(selectedFilter.length() > 0 ? " WHERE " + wherecol + " IN (" + selectedFilter + ");" : QString(";"));

	DPRINTF("query : %s", (const char*)query.utf8());
	sqlite3_exec(playListAspect()->playList()->mediaDatabase()->db(), query.toUtf8(), NULL, NULL, NULL);
}

void PlayListAspectOverviewFilter::doUpdateOnList(OverviewFilterList &list, const QString &selectedFilter, const QString &colname, const QString &srctable, const QString &unknown, const QString &one, const QString &more)
{
	//	NOTE: This query might yield incorrect results if the optimization settings at compile time are too low. Probably a bug in SQLite...
	QString query = "SELECT DISTINCT " + colname + ", " + (selectedFilter.length() > 0 ? colname + " IN (" + selectedFilter + ")" : QString("0")) +
					" AS selected FROM " + srctable + " ORDER BY " + colname + " ASC;";

	DPRINTF("query : %s", (const char*)query.toUtf8());

#ifdef QT4
	qDeleteAll(list);
#endif
	list.clear();

	sqlite3_stmt *vm;
	sqlite3_prepare_v2(playListAspect()->playList()->mediaDatabase()->db(), query.toUtf8(), -1, &vm, 0);

	OverviewFilterListEntry *entry;
	int selcount = 0;

	while (sqlite3_step(vm) == SQLITE_ROW)
	{
		entry = new OverviewFilterListEntry(
			QString::fromUtf8((const char *)sqlite3_column_text(vm, 0)),
			sqlite3_column_int(vm, 1)
		);

		if (entry->selected)
			++selcount;

		list.append(entry);
	}
	sqlite3_finalize(vm);

	entry = new OverviewFilterListEntry(
		list.count() != 1 ? more.arg(list.count()) : one,
		selcount == 0
	);
	list.insert(0, entry);

	if (list.count() > 1 && (list.at(1)->text.isEmpty()))
		list.at(1)->text = unknown;
}

void PlayListAspectOverviewFilter::updateFromGenreList()
{
	DENTERMETHOD("PlayListAspectOverviewFilter::updateFromGenreList()");

	DTIMERINIT(time);

	if (selectedGenresFilter.isEmpty())
		selectedGenresFilter = buildFilter(genreList_, tr("(Undefined genre)"));

	doUpdateOnList(genreList_, selectedGenresFilter, "genre", "overview", tr("(Undefined genre)"), tr("All (1 genre)"), tr("All (%1 genres)"));

	DTIMERPRINT(time, "updateGenres");

	updateFromArtistList();

	DEXITMETHOD("PlayListAspectOverviewFilter::updateFromGenreList()");
}

void PlayListAspectOverviewFilter::updateFromArtistList()
{
	DENTERMETHOD("PlayListAspectOverviewFilter::updateFromArtistList()");

	DTIMERINIT(time);

	selectedGenresFilter = buildFilter(genreList_, tr("(Undefined genre)"));
	createFilteredTable(selectedGenresFilter, "overview", "overview_genre_filtered", "artist, album", "genre");

	if (selectedArtistsFilter.isEmpty())
		selectedArtistsFilter = buildFilter(artistList_, tr("(Unknown artist)"));

	doUpdateOnList(artistList_, selectedArtistsFilter, "artist", "overview_genre_filtered", tr("(Unknown artist)"), tr("All (1 artist)"), tr("All (%1 artists)"));

	DTIMERPRINT(time, "updateArtists");

	updateFromAlbumList();

	DEXITMETHOD("PlayListAspectOverviewFilter::updateFromArtistList()");
}

void PlayListAspectOverviewFilter::updateFromAlbumList()
{
	DENTERMETHOD("PlayListAspectOverviewFilter::updateFromAlbumList()");

	DTIMERINIT(time);

	selectedArtistsFilter = buildFilter(artistList_, tr("(Unknown artist)"));
	createFilteredTable(selectedArtistsFilter, "overview_genre_filtered", "overview_artist_filtered", "album", "artist");

	if (selectedAlbumsFilter.isEmpty())
		selectedAlbumsFilter = buildFilter(albumList_, tr("(Unknown album)"));

	doUpdateOnList(albumList_, selectedAlbumsFilter, "album", "overview_artist_filtered", tr("(Unknown album)"), tr("All (1 album)"), tr("All (%1 albums)"));

	DTIMERPRINT(time, "updateAlbums");

	// if we're returning into the event loop because this should be interactive,
	// we need to finalize and re-prepare statements NOW, since the schema changed
	// and something may access those prepared statements in the meantime before updateView is
	// triggered by the timer...
	//if (interactive)
	//	playListView_->mediaDatabase()->ensurePreparedStatementsAreValid();
	playListAspect()->playList()->mediaDatabase()->ensurePreparedStatementsAreValid();

	// if we're entering updateView directly (i.e. non-interactively), the prepared statements will be taken care of there...
	//updateView(interactive);

	emit updateView();

	DEXITMETHOD("PlayListAspectOverviewFilter::updateFromAlbumList()");
}
